# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import Literal

import datetime as dt

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID

from gajim.common import app
from gajim.common.const import AvatarSize
from gajim.common.const import Draft
from gajim.common.const import RowHeaderType
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.contacts import ResourceContact
from gajim.common.modules.message_util import get_nickname_from_message
from gajim.common.storage.archive import models as mod
from gajim.common.types import ChatContactT
from gajim.common.util.muc import get_groupchat_name
from gajim.common.util.muc import message_needs_highlight
from gajim.common.util.preview import get_preview_data
from gajim.common.util.user_strings import get_moderation_text
from gajim.common.util.user_strings import get_retraction_text
from gajim.common.util.user_strings import get_uf_relative_time

from gajim.gtk.builder import get_builder
from gajim.gtk.menus import get_chat_list_row_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.widgets import GajimPopover


class ChatListRow(Gtk.ListBoxRow, SignalManager):
    __gsignals__ = {
        "unread-changed": (GObject.SignalFlags.RUN_LAST, None, ()),
        "context-menu-state-changed": (GObject.SignalFlags.RUN_LAST, None, (bool,)),
    }

    def __init__(
        self,
        workspace_id: str,
        account: str,
        jid: JID,
        type_: Literal["chat", "groupchat", "pm"],
        pinned: bool,
        position: int,
    ) -> None:
        Gtk.ListBoxRow.__init__(self)
        SignalManager.__init__(self)

        self.account = account
        self.jid = jid
        self.workspace_id = workspace_id
        self.type = type_
        self.position = position
        self._has_draft = False

        self._conversations_header = RowHeader(RowHeaderType.CONVERSATIONS)
        self._pinned_header = RowHeader(RowHeaderType.PINNED)

        self._client = app.get_client(account)

        self.contact = self._client.get_module("Contacts").get_contact(jid)
        assert isinstance(
            self.contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        self._connect_contact_signals()

        self.contact_name: str = self.contact.name
        self.timestamp: float = 0
        self.stanza_id: str | None = None
        self.message_id: str | None = None

        self._unread_count: int = 0
        self._needs_muc_highlight: bool = False
        self._pinned: bool = pinned

        self.add_css_class("chatlist-row")

        self._ui = get_builder("chat_list_row.ui")
        self.set_child(self._ui.mainbox)

        self._connect(self._ui.close_button, "clicked", self._on_close_button_clicked)

        self._menu_popover = GajimPopover(None)
        self._connect(self._menu_popover, "closed", self._on_context_popover_closed)
        self._ui.mainbox.append(self._menu_popover)

        self._connect(self, "state-flags-changed", self._on_state_flags_changed)

        gesture_middle_click = Gtk.GestureClick(button=Gdk.BUTTON_MIDDLE)
        self._connect(gesture_middle_click, "pressed", self._on_row_clicked)
        self.add_controller(gesture_middle_click)

        gesture_secondary_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_secondary_click, "pressed", self._on_row_clicked)
        self.add_controller(gesture_secondary_click)

        app.settings.connect_signal(
            "mute_until",
            self._on_mute_setting_changed,
            self.contact.account,
            self.contact.jid,
        )
        self._on_mute_setting_changed()

        drag_source = Gtk.DragSource(actions=Gdk.DragAction.MOVE)
        self._connect(drag_source, "prepare", self._on_prepare)
        self._connect(drag_source, "drag-begin", self._on_drag_begin)
        self._connect(drag_source, "drag-end", self._on_drag_end)
        self.add_controller(drag_source)

        if self.type == "groupchat":
            self._ui.group_chat_indicator.set_visible(True)

        self.update_avatar()
        self.update_name()
        self.update_account_identifier()

        if isinstance(self.contact, GroupchatContact) and not self.contact.can_notify():
            self._ui.unread_label.add_css_class("unread-counter-silent")

        draft = app.storage.archive.get_contact_value(
            self.contact.account, self.contact.jid, "draft"
        )
        if not self._set_draft_mode(draft):
            self._display_last_conversation_row()

    def _display_last_conversation_row(self) -> None:
        message = app.storage.archive.get_last_conversation_row(
            self.contact.account, self.contact.jid, incl_related_data=True
        )
        if message is None:
            self.set_visible(True)
            return

        assert isinstance(
            self.contact, BareContact | GroupchatContact | GroupchatParticipant
        )

        if message.text is not None:
            message_text = message.text

            if message.moderation is not None:
                message_text = get_moderation_text(
                    message.moderation.by, message.moderation.reason
                )

            elif message.retraction is not None:
                message_text = get_retraction_text(message.retraction.timestamp)

            elif corrected_message := message.get_last_correction():
                message_text = corrected_message.text
                assert message_text is not None

            # Nickname
            nickname, text_nick = get_nickname_from_message(self.account, message)
            self.set_nick(nickname)
            self.set_message_text(message_text, nickname=text_nick, oob=message.oob)

            self.set_timestamp(message.timestamp)

            self.stanza_id = message.stanza_id
            self.message_id = message.id

        if message.filetransfers:
            self.set_message_text(_("File"), icon_name="lucide-paperclip-symbolic")
            self.set_timestamp(message.timestamp)

        if message.call is not None:
            self.set_message_text(_("Call"), icon_name="lucide-phone-symbolic")
            self.set_timestamp(message.timestamp)

    @property
    def is_pinned(self) -> bool:
        return self._pinned

    @property
    def has_draft(self) -> bool:
        return self._has_draft

    @property
    def unread_count(self) -> int:
        if (
            isinstance(self.contact, GroupchatContact)
            and not self.contact.can_notify()
            and not self._needs_muc_highlight
        ):
            return 0
        return self._unread_count

    @unread_count.setter
    def unread_count(self, value: int) -> None:
        self._unread_count = value
        self._update_unread()
        self.emit("unread-changed")

    def get_header_type(self) -> RowHeaderType | None:
        header = self.get_header()
        if header is None:
            return None
        assert isinstance(header, RowHeader)
        return header.type

    def set_header_type(self, header_type: RowHeaderType | None) -> None:
        if header_type == self.get_header_type():
            return

        if header_type is None:
            self.set_header(None)

        elif header_type == RowHeaderType.PINNED:
            self.set_header(self._pinned_header)

        else:
            self.set_header(self._conversations_header)

    def clear(self) -> None:
        self.timestamp = 0
        self.stanza_id = None
        self.message_id = None
        self.set_message_text("")
        self.set_nick("")

    def set_message_id(self, message_id: str) -> None:
        self.message_id = message_id

    def set_message_text(
        self,
        text: str,
        nickname: str | None = None,
        icon_name: str | None = None,
        oob: list[mod.OOB] | None = None,
    ) -> None:
        if self._has_draft:
            return

        assert isinstance(
            self.contact, BareContact | GroupchatContact | GroupchatParticipant
        )

        icon = None
        if icon_name is not None:
            icon = Gio.Icon.new_for_string(icon_name)

        preview = get_preview_data(text, oob or [])
        if preview is not None:
            text = preview.text
            icon = preview.icon

        text = GLib.markup_escape_text(text)
        if text.startswith("/me ") and nickname is not None:
            nickname = GLib.markup_escape_text(nickname)
            text = text.replace("/me", f"* {nickname}", 1)
            text = f"<i>{text}</i>"

        # Split by newline and display last line (or first, if last is newline)
        lines = text.split("\n")
        text = lines[-1] or lines[0]

        if text.startswith("```"):
            text = _("Code snippet")

        self._ui.message_label.set_markup(text)

        if icon is None:
            self._ui.message_icon.set_visible(False)
        else:
            self._ui.message_icon.set_from_gicon(icon)
            self._ui.message_icon.set_visible(True)

    def set_nick(self, nickname: str | None) -> None:
        self._ui.nick_label.set_visible(bool(nickname))
        self._ui.nick_label.set_text(_("%(nickname)s:") % {"nickname": nickname})

    def get_real_unread_count(self) -> int:
        return self._unread_count

    def set_stanza_id(self, stanza_id: str | None) -> None:
        self.stanza_id = stanza_id

    def set_timestamp(self, timestamp: dt.datetime) -> None:
        self.timestamp = timestamp.timestamp()
        self.update_time()

    def update_account_identifier(self) -> None:
        account_class = app.css_config.get_dynamic_class(self.account)
        self._ui.account_identifier.add_css_class(account_class)
        show = len(app.settings.get_active_accounts()) > 1
        self._ui.account_identifier.set_visible(show)

    def update_avatar(self) -> None:
        scale = self.get_scale_factor()
        assert isinstance(
            self.contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        texture = self.contact.get_avatar(AvatarSize.ROSTER, scale)
        self._ui.avatar_image.set_pixel_size(AvatarSize.ROSTER)
        self._ui.avatar_image.set_from_paintable(texture)

    def update_name(self) -> None:
        assert isinstance(
            self.contact, BareContact | GroupchatContact | GroupchatParticipant
        )
        if self.type == "pm":
            client = app.get_client(self.account)
            muc_name = get_groupchat_name(client, self.jid.new_as_bare())
            self._ui.name_label.set_text(f"{self.contact.name} ({muc_name})")
            return

        self.contact_name = self.contact.name
        if self.jid == self._client.get_own_jid().bare:
            self.contact_name = _("Note to myself")
        self._ui.name_label.set_text(self.contact_name)

    def update_time(self) -> None:
        if self.timestamp == 0:
            return
        utc_timestamp = dt.datetime.fromtimestamp(self.timestamp, dt.UTC)
        self._ui.timestamp_label.set_text(get_uf_relative_time(utc_timestamp))

    def update_row_state(self) -> None:
        self.update_time()
        self._ui.mute_image.set_visible(self.contact.is_muted)

    def add_unread(self, text: str) -> None:
        assert self.message_id is not None
        self._unread_count += 1
        self._update_unread()
        app.storage.cache.set_unread_count(
            self.account,
            self.jid,
            self.get_real_unread_count(),
            self.message_id,
            self.timestamp,
        )

        if (
            isinstance(self.contact, GroupchatContact)
            and self.contact.nickname is not None
        ):
            if message_needs_highlight(
                text, self.contact.nickname, self._client.get_own_jid().bare
            ):
                self._needs_muc_highlight = True
                self._ui.mention_indicator.set_visible(True)

        self.emit("unread-changed")

    def reset_unread(self) -> None:
        self._needs_muc_highlight = False
        self._ui.mention_indicator.set_visible(False)

        self._unread_count = 0
        self._update_unread()
        self.emit("unread-changed")

        app.storage.cache.reset_unread_count(self.account, self.jid)

    def toggle_pinned(self) -> None:
        self._pinned = not self._pinned

    def _update_unread(self) -> None:
        unread_count = self._get_unread_string(self._unread_count)
        self._ui.unread_label.set_text(unread_count)
        self._ui.unread_label.set_visible(bool(self._unread_count))

    @staticmethod
    def _get_unread_string(count: int) -> str:
        if count < 1000:
            return str(count)
        return "999+"

    def _on_mute_setting_changed(self, *args: Any) -> None:
        self._ui.mute_image.set_visible(self.contact.is_muted)

    def _on_draft_update(
        self,
        _contact: ChatContactT,
        _signal_name: str,
        draft: Draft | None,
    ) -> None:
        if not self._set_draft_mode(draft):
            self._display_last_conversation_row()
        self.changed()

    def _set_draft_mode(self, draft: Draft | None) -> bool:
        self._has_draft = draft is not None

        if draft is None:
            self._ui.message_label.remove_css_class("draft")

        else:
            self.set_nick("")
            self._ui.message_label.set_text(_("Draft: %s") % draft.text)
            self._ui.message_label.add_css_class("draft")

        return self._has_draft

    def _on_state_flags_changed(
        self, _row: ChatListRow, _flags: Gtk.StateFlags
    ) -> None:
        state = self.get_state_flags()
        if (state & Gtk.StateFlags.PRELIGHT) != 0:
            self._ui.revealer.set_reveal_child(True)
        else:
            self._ui.revealer.set_reveal_child(False)

    def do_unroot(self) -> None:
        self._disconnect_all()

        app.settings.disconnect_signals(self)

        self._client.disconnect_all_from_obj(self)
        self.contact.disconnect_all_from_obj(self)
        if isinstance(self.contact, GroupchatParticipant):
            self.contact.room.disconnect_all_from_obj(self)

        del self._menu_popover
        Gtk.ListBoxRow.do_unroot(self)
        app.check_finalize(self)

    def _on_close_button_clicked(self, _button: Gtk.Button) -> None:
        self.reset_unread()
        app.window.activate_action(
            "win.remove-chat", GLib.Variant("as", [self.account, str(self.jid)])
        )

    def _on_row_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        if gesture_click.get_current_button() == Gdk.BUTTON_MIDDLE:
            self.reset_unread()
            app.window.activate_action(
                "win.remove-chat", GLib.Variant("as", [self.account, str(self.jid)])
            )
        elif gesture_click.get_current_button() == Gdk.BUTTON_SECONDARY:
            self._raise_context_popover(x, y)

    def _raise_context_popover(self, x: float, y: float):
        menu = get_chat_list_row_menu(
            self.workspace_id, self.account, self.jid, self._pinned
        )

        self._menu_popover.set_menu_model(menu)
        self._menu_popover.set_pointing_to_coord(x=x, y=y)
        self.emit("context-menu-state-changed", True)
        self._menu_popover.popup()

    def _on_context_popover_closed(self, _popover: Gtk.Popover) -> None:
        self.emit("context-menu-state-changed", False)

    def _on_prepare(
        self, _drag_source: Gtk.DragSource, _x: float, _y: float
    ) -> Gdk.ContentProvider:
        value = GObject.Value()
        value.init(ChatListRow)
        value.set_object(self)

        return Gdk.ContentProvider.new_for_value(value)

    def _on_drag_begin(self, _drag_source: Gtk.DragSource, drag: Gdk.Drag) -> None:
        assert not isinstance(self.contact, ResourceContact)

        is_LTR = bool(self.get_direction() == Gtk.TextDirection.LTR)

        Gtk.DragIcon.set_from_paintable(
            drag,
            Gtk.WidgetPaintable().new(self),
            hot_x=0 if is_LTR else self.get_width(),
            hot_y=0,
        )

        app.window.highlight_dnd_targets(self, True)

    def _on_drag_end(
        self, _drag_source: Gtk.DragSource, _drag: Gdk.Drag, _delete_data: bool
    ) -> None:
        app.window.highlight_dnd_targets(self, False)

    def _connect_contact_signals(self) -> None:
        self.contact.connect("chatstate-update", self._on_chatstate_update)
        self.contact.connect("draft-update", self._on_draft_update)

        if isinstance(self.contact, BareContact):
            self.contact.connect("presence-update", self._on_presence_update)
            self.contact.connect("nickname-update", self._on_name_update)
            self.contact.connect("caps-update", self._on_avatar_update)
            self.contact.connect("avatar-update", self._on_avatar_update)

        elif isinstance(self.contact, GroupchatContact):
            self.contact.connect("avatar-update", self._on_avatar_update)
            self.contact.connect("name-update", self._on_name_update)
            self.contact.connect("disco-info-update", self._on_disco_info_update)
            self.contact.connect("state-changed", self._on_muc_state_changed)
            self.contact.connect("mam-sync-started", self._on_mam_sync_changed)
            self.contact.connect("mam-sync-finished", self._on_mam_sync_changed)
            self.contact.connect("mam-sync-error", self._on_mam_sync_error)

            self._client.connect_signal("state-changed", self._on_client_state_changed)

        elif isinstance(self.contact, GroupchatParticipant):
            self.contact.connect("user-joined", self._on_muc_user_update)
            self.contact.connect("user-left", self._on_muc_user_update)
            self.contact.connect("user-avatar-update", self._on_muc_user_update)
            self.contact.connect("user-status-show-changed", self._on_muc_user_update)

            self.contact.room.connect("room-left", self._on_muc_update)
            self.contact.room.connect("room-destroyed", self._on_muc_update)
            self.contact.room.connect("room-kicked", self._on_muc_update)

        else:
            raise TypeError("Unknown contact type: %s" % type(self.contact))

    def _on_presence_update(self, _contact: ChatContactT, _signal_name: str) -> None:
        self.update_avatar()

    def _on_avatar_update(self, _contact: ChatContactT, _signal_name: str) -> None:
        self.update_avatar()

    def _on_disco_info_update(self, _contact: ChatContactT, _signal_name: str) -> None:
        self.update_name()

    def _on_client_state_changed(self, *args: Any) -> None:
        self._update_joined_state()

    def _on_muc_state_changed(self, *args: Any) -> None:
        self._update_joined_state()

    def _update_joined_state(self) -> None:
        self._reset_connection_icon()

        assert isinstance(self.contact, GroupchatContact)
        if self.contact.is_joining:
            self._ui.connection_icon.set_from_icon_name("lucide-refresh-cw-symbolic")
            self._ui.connection_icon.add_css_class("spin")
            self._ui.connection_icon.add_css_class("dimmed")
            self._ui.connection_icon.set_tooltip_text(_("Joining Group Chat…"))
            self._ui.connection_icon.set_visible(True)
        elif self.contact.is_not_joined or not self._client.is_available():
            self._ui.connection_icon.set_from_icon_name("lucide-zap-symbolic")
            self._ui.connection_icon.add_css_class("warning")
            self._ui.connection_icon.set_tooltip_text(_("Not connected"))
            self._ui.connection_icon.set_visible(True)

    def _on_mam_sync_changed(
        self, _contact: GroupchatContact, signal_name: str
    ) -> None:
        self._reset_connection_icon()

        if signal_name == "mam-sync-started":
            self._ui.connection_icon.set_from_icon_name("lucide-refresh-cw-symbolic")
            self._ui.connection_icon.add_css_class("spin")
            self._ui.connection_icon.add_css_class("accent")
            self._ui.connection_icon.set_tooltip_text(_("Fetching messages…"))
            self._ui.connection_icon.set_visible(True)

    def _on_mam_sync_error(
        self, _contact: GroupchatContact, _signal_name: str, error_text: str
    ) -> None:
        self._reset_connection_icon()
        self._ui.connection_icon.set_from_icon_name("lucide-zap-symbolic")
        self._ui.connection_icon.add_css_class("error")
        self._ui.connection_icon.set_tooltip_text(
            _("There has been an error while trying to fetch messages: %s") % error_text
        )
        self._ui.connection_icon.set_visible(True)

    def _reset_connection_icon(self) -> None:
        self._ui.connection_icon.set_visible(False)
        self._ui.connection_icon.remove_css_class("spin")
        self._ui.connection_icon.remove_css_class("dimmed")
        self._ui.connection_icon.remove_css_class("accent")
        self._ui.connection_icon.remove_css_class("warning")
        self._ui.connection_icon.remove_css_class("error")

    def _on_muc_user_update(
        self, _contact: GroupchatParticipant, _signal_name: str, *args: Any
    ) -> None:
        self.update_avatar()

    def _on_muc_update(
        self, _contact: GroupchatContact, _signal_name: str, *args: Any
    ) -> None:
        self.update_avatar()

    def _on_chatstate_update(self, contact: ChatContactT, _signal_name: str) -> None:
        if contact.is_groupchat:
            assert isinstance(contact, GroupchatContact)
            self._ui.chatstate_image.set_visible(contact.has_composing_participants())
            return

        if contact.chatstate is None:
            self._ui.chatstate_image.set_visible(False)
        else:
            self._ui.chatstate_image.set_visible(contact.chatstate.is_composing)

    def _on_name_update(self, _contact: ChatContactT, _signal_name: str) -> None:
        self.update_name()


class RowHeader(Gtk.Box):
    def __init__(self, header_type: RowHeaderType) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.type = header_type

        if header_type == RowHeaderType.CONVERSATIONS:
            text = _("Conversations")
        else:
            text = _("Pinned")

        label = Gtk.Label(label=text)
        label.set_halign(Gtk.Align.START)

        self.append(label)

        self.add_css_class("header-box")
        if header_type == RowHeaderType.PINNED:
            self.add_css_class("header-box-first")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0202: Entity Time

from __future__ import annotations

from typing import cast

from nbxmpp.protocol import JID

from gajim.common import types
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact


class EntityTime(BaseModule):
    _nbxmpp_extends = "EntityTime"
    _nbxmpp_methods = [
        "request_entity_time",
        "enable",
        "disable",
    ]

    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = []

    def set_enabled(self, enabled: bool) -> None:
        if not enabled:
            self._nbxmpp("EntityTime").disable()
            return

        self._nbxmpp("EntityTime").enable()
        self._nbxmpp("EntityTime").set_allow_reply_func(self._allow_reply)

    def _allow_reply(self, jid: JID) -> bool:
        bare_jid = jid.new_as_bare()
        item = self._con.get_module("Roster").get_item(bare_jid)
        if item is None:
            return False

        if not self._client.get_module("VCard4").is_timezone_published():
            self._log.info(
                "Ignore entity time request, because timezone is not published"
            )
            return False

        contact = cast(BareContact, self._get_contact(bare_jid))
        return contact.is_subscribed



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import dataclasses
import datetime as dt

import nbxmpp
import sqlalchemy.exc
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import JID
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler
from nbxmpp.util import generate_id

from gajim.common import app
from gajim.common import types
from gajim.common.const import RETRACTION_FALLBACK
from gajim.common.const import Trust
from gajim.common.events import MessageAcknowledged
from gajim.common.events import MessageCorrected
from gajim.common.events import MessageError
from gajim.common.events import MessageReceived
from gajim.common.events import MessageRetracted
from gajim.common.events import MessageSent
from gajim.common.events import ReactionUpdated
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.message_util import convert_message_type
from gajim.common.modules.message_util import get_chat_type_and_direction
from gajim.common.modules.message_util import get_eme_message
from gajim.common.modules.message_util import get_eme_protocol
from gajim.common.modules.message_util import get_message_timestamp
from gajim.common.modules.message_util import get_occupant_info
from gajim.common.modules.message_util import get_open_graph
from gajim.common.modules.message_util import get_reply
from gajim.common.modules.message_util import get_security_label
from gajim.common.modules.misc import parse_oob
from gajim.common.storage.archive import models as mod
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageState
from gajim.common.storage.archive.const import MessageType
from gajim.common.structs import OutgoingMessage


class Message(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="message", callback=self._check_if_unknown_contact, priority=41
            ),
            StanzaHandler(name="message", callback=self._message_received, priority=50),
            StanzaHandler(
                name="message",
                typ="error",
                callback=self._message_error_received,
                priority=50,
            ),
        ]

        # XEPs for which this message module should not be executed
        self._message_namespaces = {Namespace.ROSTERX, Namespace.IBB}

    def _check_if_unknown_contact(
        self,
        _con: types.NBXMPPClient,
        stanza: nbxmpp.Message,
        properties: MessageProperties,
    ) -> None:
        if (
            properties.type.is_groupchat
            or properties.is_muc_pm
            or properties.is_self_message
            or properties.is_mam_message
        ):
            return

        if self._con.get_own_jid().domain == str(properties.jid):
            # Server message
            return

        if not app.settings.get_account_setting(
            self._account, "ignore_unknown_contacts"
        ):
            return

        assert properties.jid is not None
        bare_jid = properties.jid.new_as_bare()
        if self._con.get_module("Roster").get_item(bare_jid) is None:
            self._log.warning("Ignore message from unknown contact: %s", bare_jid)
            self._log.warning(stanza)
            raise nbxmpp.NodeProcessed

    def _message_received(
        self,
        _con: types.NBXMPPClient,
        stanza: nbxmpp.Message,
        properties: MessageProperties,
    ) -> None:

        if properties.is_pubsub or properties.type.is_error:
            return

        if properties.type.is_normal and not properties.body:
            return

        # Check if a child of the message contains any
        # namespaces that we handle in other modules.
        # nbxmpp executes less common handlers last
        if self._message_namespaces & set(stanza.getProperties()):
            return

        self._log.info("Received from %s", stanza.getFrom())

        if properties.carbon is not None and properties.carbon.is_sent:
            # Ugly, we treat the from attr as the remote jid,
            # to make that work with sent carbons we have to do this.
            # TODO: Check where in Gajim and plugins we depend on that behavior
            to = stanza.getTo()
            assert to is not None
            stanza.setFrom(to)

        timestamp = get_message_timestamp(properties)
        remote_jid = properties.remote_jid
        jid = properties.jid
        assert remote_jid is not None
        assert jid is not None

        muc_data = None
        if properties.type.is_groupchat:
            muc_data = self._client.get_module("MUC").get_muc_data(remote_jid)
            if muc_data is None:
                self._log.warning("Groupchat message from unknown MUC: %s", remote_jid)
                return

        m_type, direction = get_chat_type_and_direction(
            muc_data, self._client.get_own_jid(), properties
        )

        user_delay_ts = None
        if properties.user_timestamp is not None:
            user_delay_ts = dt.datetime.fromtimestamp(
                properties.user_timestamp, tz=dt.UTC
            )

        message_id = properties.id
        if message_id is None:
            # TODO: Make Gajim not depend on a message_id being present
            message_id = generate_id()
            self._log.debug("Generating id for message")

        stanza_id = self._get_stanza_id(properties)

        # Fallback to message-id in case the MUC strips origin-id
        # https://dev.gajim.org/gajim/gajim/-/issues/11837
        origin_id = properties.origin_id or properties.id

        if (
            m_type in (MessageType.CHAT, MessageType.PM)
            and direction == ChatDirection.OUTGOING
            and origin_id is not None
        ):
            if app.storage.archive.check_if_message_id_exists(
                self._account, remote_jid, origin_id
            ):
                self._log.info("Duplicated message received: %s", origin_id)
                return

        if (
            m_type == MessageType.GROUPCHAT
            and direction == ChatDirection.OUTGOING
            and origin_id is not None
        ):
            # Use origin-id because some group chats change the message id
            # on the reflection.
            pk = app.storage.archive.update_pending_message(
                self._account, remote_jid, origin_id, stanza_id
            )

            if pk is not None:
                app.ged.raise_event(
                    MessageAcknowledged(
                        account=self._account,
                        jid=remote_jid,
                        pk=pk,
                        stanza_id=stanza_id,
                    )
                )
                return

        resource = jid.resource
        if m_type == MessageType.PM and direction == ChatDirection.OUTGOING:
            # Outgoing PMs which we receive as carbons or via MAM
            # have our bound jid in the from attribute, not the joined MUC
            # resource. This means for these kind of messages the nickname
            # cannot be determined
            resource = None

        occupant = None
        if m_type in (MessageType.GROUPCHAT, MessageType.PM) and not jid.is_bare:
            contact = self._client.get_module("Contacts").get_contact(
                jid, groupchat=True
            )
            assert isinstance(contact, GroupchatParticipant)

            occupant = get_occupant_info(
                self._account,
                remote_jid,
                self._get_own_bare_jid(),
                direction,
                m_type,
                timestamp,
                contact,
                properties,
            )

        assert properties.bodies is not None
        message_text = properties.bodies.get(None)
        oob_data = parse_oob(properties)

        encryption_data = None
        if properties.encrypted is not None:
            encryption_data = mod.Encryption(**dataclasses.asdict(properties.encrypted))

        elif properties.eme is not None:
            message_text = get_eme_message(properties.eme)
            protocol = get_eme_protocol(properties.eme)
            encryption_data = mod.Encryption(
                protocol=protocol, key="Unknown", trust=Trust.UNTRUSTED
            )

        if not message_text:
            self._log.debug("Received message without text")
            return

        securitylabel_data = get_security_label(
            self._account, remote_jid, timestamp, properties.security_label
        )
        reply = get_reply(properties.reply_data)

        correction_id = None
        if properties.correction is not None:
            correction_id = properties.correction.id

        og_data = get_open_graph(properties.open_graph)

        message_data = mod.Message(
            account_=self._account,
            remote_jid_=remote_jid,
            type=m_type,
            direction=direction,
            timestamp=timestamp,
            state=MessageState.ACKNOWLEDGED,
            resource=resource,
            text=message_text,
            id=message_id,
            stanza_id=stanza_id,
            user_delay_ts=user_delay_ts,
            correction_id=correction_id,
            encryption_=encryption_data,
            occupant_=occupant,
            oob=oob_data,
            security_label_=securitylabel_data,
            reply=reply,
            thread_id_=properties.thread,
            og=og_data,
        )

        try:
            pk = app.storage.archive.insert_object(
                message_data, ignore_on_conflict=False
            )
        except sqlalchemy.exc.IntegrityError:
            self._log.exception("Insertion Error")
            return

        if correction_id is not None:
            event = MessageCorrected(
                account=self._account,
                jid=remote_jid,
                pk=pk,
                correction_id=correction_id,
            )
            app.ged.raise_event(event)
            return

        app.ged.raise_event(
            MessageReceived(
                account=self._account,
                jid=remote_jid,
                m_type=m_type,
                mam=properties.mam,
                pk=pk,
            )
        )

    def _message_error_received(
        self,
        _con: types.NBXMPPClient,
        stanza: nbxmpp.Message,
        properties: MessageProperties,
    ) -> None:

        remote_jid = properties.remote_jid
        assert remote_jid is not None
        message_id = properties.id
        if message_id is None:
            self._log.warning("Received error without message id")
            self._log.warning(stanza)
            return

        timestamp = get_message_timestamp(properties)

        assert properties.error is not None
        assert properties.jid is not None

        # Some servers do not include the original payload in the error
        # This makes it hard to detect MUC PMs. Try to find the corresponding
        # message in the database to narrow it down.
        remotes = [properties.jid]
        if not properties.jid.is_bare:
            remotes.append(properties.jid.new_as_bare())

        message = app.storage.archive.get_message_from_error_id(
            self._account, remotes, message_id
        )

        if message is not None:
            remote_jid = message.remote.jid

        error_data = mod.MessageError(
            account_=self._account,
            remote_jid_=remote_jid,
            message_id=message_id,
            by=properties.error.by,
            type=properties.error.type,
            text=properties.error.get_text() or None,
            condition=properties.error.condition,
            condition_text=properties.error.condition_data,
            timestamp=timestamp,
        )

        pk = app.storage.archive.insert_row(error_data, ignore_on_conflict=True)
        if pk == -1:
            self._log.warning(
                "Received error with already known message id: %s", message_id
            )
            return

        app.ged.raise_event(
            MessageError(
                account=self._account,
                jid=remote_jid,
                message_id=message_id,
                error=properties.error,
            )
        )

    def _get_stanza_id(self, properties: MessageProperties) -> str | None:

        if properties.is_mam_message:
            assert properties.mam is not None
            return properties.mam.id

        if not properties.stanza_ids:
            return None

        if properties.type.is_groupchat:
            archive = properties.remote_jid
            assert archive is not None
            disco_info = app.storage.cache.get_last_disco_info(archive)
            assert disco_info is not None
            if not disco_info.supports(Namespace.SID):
                return None

        else:
            if not self._con.get_module("MAM").available:
                return None

            archive = self._con.get_own_jid().new_as_bare()

        for stanza_id in properties.stanza_ids:
            # Check if message is from expected archive
            if archive.bare_match(stanza_id.by):
                return stanza_id.id
        return None

    def store_message(self, message: OutgoingMessage) -> None:
        if (
            not message.has_text()
            and message.reaction_data is None
            and message.retraction_id is None
        ):
            return

        if message.type == MessageType.GROUPCHAT:
            if message.reaction_data is not None or message.retraction_id is not None:
                # Store reaction when the MUC reflects it
                return

        direction = ChatDirection.OUTGOING
        remote_jid = message.contact.jid

        assert message.message_id is not None

        occupant = None
        resource = self._client.get_bound_jid().resource
        state = MessageState.ACKNOWLEDGED

        if message.type in (MessageType.GROUPCHAT, MessageType.PM):
            # PM is a full jid, so convert to bare
            muc_jid = remote_jid.new_as_bare()
            muc_data = self._client.get_module("MUC").get_muc_data(muc_jid)
            if muc_data is None:
                self._log.warning("Trying to send message to unknown MUC: %s", muc_jid)
                return

            resource = muc_data.nick
            real_jid = self._client.get_own_jid().new_as_bare()
            if message.type == MessageType.PM:
                # Use real jid as occupant-id because sent MUC PMs which we
                # receive via carbons or MAM do not have the occupant-id
                # which would lead to outgoing message being stored with
                # different occupant-ids
                occupant_id = real_jid
            else:
                occupant_id = muc_data.occupant_id or real_jid

            occupant = mod.Occupant(
                account_=self._account,
                remote_jid_=muc_jid,
                id=str(occupant_id),
                real_remote_jid_=real_jid,
                nickname=resource,
                updated_at=message.timestamp,
            )

            if message.type == MessageType.GROUPCHAT:
                state = MessageState.PENDING

        if message.reaction_data is not None:
            reactions_id, reactions = message.reaction_data
            if not reactions:
                app.storage.archive.delete_reaction(
                    account=self._account,
                    jid=remote_jid,
                    occupant_id=occupant.id if occupant is not None else None,
                    reaction_id=reactions_id,
                    direction=direction,
                )
                self._log.info("Delete own reactions for: %s", reactions_id)

            else:
                reaction = mod.Reaction(
                    account_=self._account,
                    remote_jid_=remote_jid,
                    occupant_=occupant,
                    id=reactions_id,
                    direction=direction,
                    emojis=";".join(reactions),
                    timestamp=message.timestamp,
                )

                app.storage.archive.upsert_row2(reaction, return_full=False)

            app.ged.raise_event(
                ReactionUpdated(
                    account=self._account,
                    jid=remote_jid,
                    id=reactions_id,
                    direction=direction,
                    occupant=occupant,
                    emojis=None,
                    message=None,
                    is_mam_message=False,
                )
            )
            return

        if message.retraction_id is not None:
            retraction = mod.Retraction(
                account_=self._account,
                remote_jid_=remote_jid,
                occupant_=occupant,
                id=message.retraction_id,
                direction=direction,
                timestamp=message.timestamp,
            )

            app.storage.archive.insert_row(retraction, ignore_on_conflict=True)

            app.ged.raise_event(
                MessageRetracted(
                    account=self._account, jid=remote_jid, retraction=retraction
                )
            )
            return

        encryption_data = None
        encryption = message.get_encryption()
        if encryption is not None:
            encryption_data = mod.Encryption(**dataclasses.asdict(encryption))

        securitylabel_data = get_security_label(
            self._account, remote_jid, message.timestamp, message.sec_label
        )
        reply = get_reply(message.reply_data)

        thread_id = None
        if message.reply_data is not None:
            thread_id = message.reply_data.thread_id

        oob_data: list[mod.OOB] = []
        if message.oob_url is not None:
            oob_data.append(mod.OOB(url=message.oob_url, description=None))

        open_graph_data: list[mod.OpenGraph] = []
        if message.open_graph_data:
            open_graph_data = get_open_graph(message.open_graph_data)

        message_data = mod.Message(
            account_=self._account,
            remote_jid_=remote_jid,
            type=message.type,
            direction=direction,
            timestamp=message.timestamp,
            state=state,
            resource=resource,
            text=message.get_text(with_fallback=False),
            id=message.message_id,
            stanza_id=None,
            user_delay_ts=None,
            thread_id_=thread_id,
            correction_id=message.correct_id,
            encryption_=encryption_data,
            oob=oob_data,
            og=open_graph_data,
            reply=reply,
            security_label_=securitylabel_data,
            occupant_=occupant,
        )

        pk = app.storage.archive.insert_object(message_data)
        if pk == -1:
            return

        if message.correct_id is not None:
            event = MessageCorrected(
                account=self._account,
                jid=remote_jid,
                pk=pk,
                correction_id=message.correct_id,
            )
            app.ged.raise_event(event)
            return

        app.ged.raise_event(
            MessageSent(
                jid=remote_jid,
                account=message.account,
                pk=pk,
                play_sound=message.play_sound,
            )
        )


def build_message_stanza(message: OutgoingMessage, own_jid: JID) -> nbxmpp.Message:
    remote_jid = message.contact.jid

    stanza = nbxmpp.Message(
        to=str(remote_jid),
        body=message.get_text(),
        typ=convert_message_type(message.type),
    )

    # XEP-0359
    stanza.setID(message.message_id)
    stanza.setOriginID(message.message_id)

    thread_id = None

    # Mark Message as MUC PM
    if message.is_pm:
        stanza.setTag("x", namespace=Namespace.MUC_USER)

    # XEP-0444
    if message.reaction_data is not None:
        stanza.setReactions(*message.reaction_data)
        stanza.setTag("store", namespace=Namespace.MSG_HINTS)
        return stanza

    if message.correct_id:
        stanza.setTag(
            "replace", attrs={"id": message.correct_id}, namespace=Namespace.CORRECT
        )

    # XEP-0461
    if message.reply_data is not None:
        thread_id = message.reply_data.thread_id
        stanza.setReply(
            str(message.reply_data.to),
            message.reply_data.id,
            message.reply_data.fallback_start,
            message.reply_data.fallback_end,
        )

    # OGP link previews
    if message.open_graph_data is not None:
        for about_url, open_graph_data in message.open_graph_data.items():
            stanza.addOpenGraph(about_url, open_graph_data.to_nbxmpp())

    # XEP-0258
    if message.sec_label:
        stanza.addChild(node=message.sec_label.to_node())

    # XEP-0066
    if message.oob_url is not None:
        oob = stanza.addChild("x", namespace=Namespace.X_OOB)
        oob.addChild("url").setData(message.oob_url)

    # XEP-0184
    if not own_jid.bare_match(message.contact.jid):
        if message.has_text() and not message.is_groupchat:
            stanza.setReceiptRequest()

    # XEP-0085
    if message.chatstate is not None:
        stanza.setTag(message.chatstate, namespace=Namespace.CHATSTATES)
        if not message.has_text():
            stanza.setTag("no-store", namespace=Namespace.MSG_HINTS)

    # XEP-0333
    if message.has_text():
        stanza.setMarkable()
    if message.marker:
        marker, id_ = message.marker
        stanza.setMarker(marker, id_)

    # XEP-0424
    if message.retraction_id is not None:
        stanza.setRetracted(message.retraction_id, fallback_text=RETRACTION_FALLBACK)

    # XEP-0490
    if message.mds_id is not None:
        stanza.setMdsAssist(message.mds_id, own_jid.new_as_bare())

    # Thread
    # Currently only supported for replies
    if thread_id is not None:
        stanza.setThread(thread_id)

    return stanza



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common.i18n import _

from gajim.gtk.const import Setting
from gajim.gtk.const import SettingKind
from gajim.gtk.const import SettingType
from gajim.gtk.settings import GajimPreferencesGroup


class ContactSettings(GajimPreferencesGroup):
    def __init__(self, account: str, jid: JID) -> None:
        GajimPreferencesGroup.__init__(
            self,
            key="main",
            account=account,
            jid=str(jid),
            title=_("Privacy"),
        )

        chat_state = {
            "disabled": _("Disabled"),
            "composing_only": _("Composing Only"),
            "all": _("All Chat States"),
        }

        settings: list[Setting] = [
            Setting(
                SettingKind.DROPDOWN,
                _("Send Chat State"),
                SettingType.CONTACT,
                "send_chatstate",
                props={"data": chat_state},
            ),
            Setting(
                SettingKind.SWITCH,
                _("Read Receipts"),
                SettingType.CONTACT,
                "send_marker",
                desc=_("Send and receive read receipts"),
            ),
            Setting(
                SettingKind.SWITCH,
                _("Show Link Preview"),
                SettingType.CONTACT,
                "enable_link_preview",
                callback=app.window.reload_view,
            ),
        ]

        for setting in settings:
            self.add_setting(setting)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0280: Message Carbons

from __future__ import annotations

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.structs import DiscoInfo

from gajim.common import types
from gajim.common.modules.base import BaseModule


class Carbons(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.supported = False

    def pass_disco(self, info: DiscoInfo) -> None:
        if Namespace.CARBONS not in info.features:
            return

        self.supported = True
        self._log.info("Discovered carbons: %s", info.jid)

        iq = nbxmpp.Iq("set")
        iq.setTag("enable", namespace=Namespace.CARBONS)
        self._log.info("Activate")
        self._con.connection.send(iq)
